<?php
/**
 * Copyright (C) 2019 thirty bees
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Academic Free License (AFL 3.0)
 * that is bundled with this package in the file LICENSE.md.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/afl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@thirtybees.com so we can send you a copy immediately.
 *
 * @author    thirty bees <modules@thirtybees.com>
 * @copyright 2019 thirty bees
 * @license   Academic Free License (AFL 3.0)
 */

if (!defined('_TB_VERSION_')) {
    exit;
}

/**
 * Class GitUpdate.
 *
 * This class handles collected knowledge about a given update or version
 * comparison. It's designed to collect this knowledge in small steps, to give
 * merchants a good GUI feedback and avoid running into maximum execution time
 * limits.
 */
class GitUpdate
{
    const STORAGE_PATH = _PS_CACHE_DIR_.'/GitUpdateStorage.php';
    /**
     * Storage items NOT changing when choosing a different original/target
     * version. Not deleting them speeds up comparison for merchants dialing
     * through several versions.
     */
    const STORAGE_PERMANENT_ITEMS = [
        '#^filelist-[0-9\.]+$#',
    ];

    /**
     * Set of regular expressions for removing file paths from the list of
     * files of a full release package. Matching paths get ignored by
     * comparions and by updates.
     */
    const RELEASE_PATH_FILTER = [
        '#^install/#',
        '#^modules/#',
    ];
    /**
     * Set of regular expressions for removing file paths from the list of
     * local files. Ignored for files in either the original or the target
     * release.
     */
    const INSTALLATION_PATH_FILTER = [
        '#^cache/#',
        '#^img/#',
        '#^upload/#',
        '#^download/#',
        '#^translations/#',
        '#^mails/#',
        '#^override/#',
        '#^themes/(?!community-theme-default/)#',
        '#^themes/community-theme-default/cache/#',
        '#^themes/community-theme-default/lang/#',
        '#^themes/community-theme-default/mails/#',
        '#^config/settings.inc.php$#',
        '#^config/settings.old.php$#',
        '#^config/defines_custom.inc.php$#',
        '#^.htaccess$#',
        '#^robots.txt$#',
    ];
    /**
     * These files are left untouched even if they come with one of the
     * releases. All these files shouldn't be distributed in this location, to
     * begin with, but copied there from install/ at installation time.
     */
    const KEEP_PATH_FILTER = [
        '#^img/favicon.ico$#',
        '#^img/favicon_[0-9]+$#',
        '#^img/logo.jpg$#',
        '#^img/logo_stores.png$#',
        '#^img/logo_invoice.jpg$#',
        '#^img/c/[0-9-]+_thumb.jpg$#',
        '#^img/s/[0-9]+.jpg$#',
        '#^img/t/[0-9]+.jpg$#',
        '#^img/cms/cms-img.jpg$#',
    ];

    /**
     * @var GitUpdate
     */
    public static $instance = null;

    /**
     * The signature prohibits instantiating a non-singleton class.
     *
     * @since 1.0.0
     */
    protected function __construct() { }

    /**
     * Do one step to build a comparison, starting with the first step.
     * Subsequent calls see results of the previous step and proceed with the
     * next one accordingly.
     *
     * @param $messages Prepared array to append messages to. Format see
     *                  AdminGitUpdaterController->ajaxProcessCompare().
     * @param $version  Version to compare the installation on disk to.
     *
     * @since 1.0.0
     */
    public static function compareStep(&$messages, $version)
    {
        static $storage = [];

        // Maintain a singleton instance to allow re-use of network
        // connections and similar stuff.
        if ( ! static::$instance) {
            static::$instance = new static();
        }
        $me = static::$instance;

        // Dump very old storage.
        if (file_exists(static::STORAGE_PATH)
            && time() - filemtime(static::STORAGE_PATH) > 86400) {
            static::deleteStorage(true);
        }

        // Load storage.
        if ( ! count($storage) && is_readable(static::STORAGE_PATH)) {
            require static::STORAGE_PATH;
        }

        // Reset an invalid storage set.
        if ( ! array_key_exists('versionOrigin', $storage)
            || $storage['versionOrigin'] !== _TB_VERSION_
            || ! array_key_exists('versionTarget', $storage)
            || $storage['versionTarget'] !== $version) {

            static::deleteNonpermanentStorageItems($storage);

            $storage['versionOrigin'] = _TB_VERSION_;
            $storage['versionTarget'] = $version;
        }

        // Do one compare step.
        if ( ! array_key_exists('filelist-'.$version, $storage)) {
            $fileList = $me->downloadFileList($version);
            if ($fileList !== false) {
                $storage['filelist-'.$version] = $fileList;

                $messages['informations'][] =
                    sprintf($me->l('File list for version %s downloaded.'), $version)
                    .' '.sprintf($me->l('Found %d paths to consider.'), count($fileList));
                $messages['done'] = false;
            } else {
                $messages['informations'][] = sprintf($me->l('Failed to download file list for version %s.'), $version);
                $messages['error'] = true;
            }
        } elseif ( ! array_key_exists('filelist-'._TB_VERSION_, $storage)) {
            $fileList = $me->downloadFileList(_TB_VERSION_);
            if ($fileList !== false) {
                $storage['filelist-'._TB_VERSION_] = $fileList;

                $messages['informations'][] =
                    sprintf($me->l('File list for version %s downloaded.'), _TB_VERSION_);
                $messages['done'] = false;
            } else {
                $messages['informations'][] = sprintf($me->l('Failed to download file list for version %s.'), _TB_VERSION_);
                $messages['error'] = true;
            }
        } elseif ( ! array_key_exists('toplevel-'.$version, $storage)) {
            $topLevelDirs = $me->topLevelDirs($storage['filelist-'.$version]);
            $storage['toplevel-'.$version] = $topLevelDirs;
            $storage['installationlist'] = [];

            $messages['informations'][] = $me->l('Extracted top level directories.');
            $messages['done'] = false;
        } elseif (count($storage['toplevel-'.$version])) {
            $dir = array_pop($storage['toplevel-'.$version]);
            $me->searchInstallation($storage, $dir);

            $messages['informations'][] = sprintf($me->l('Searched installed files in %s/'), $dir);
            $messages['done'] = false;
        } else {
            $me->calculateChanges($storage);
            $messages['changeset'] = $storage['changeset'];

            $messages['informations'][] = $me->l('Changeset calculated. Done.');
            $messages['done'] = true;
        }

        static::saveStorage($storage);

        // @TODO: remove this when using storage for actual updates.
        if ($messages['done']) {
            static::deleteStorage(false);
        }
    }

    /**
     * Save storage. File format allows to PHP include the resulting file.
     *
     * @param mixed $storage Variable to store.
     *
     * @since 1.0.0
     */
    public static function saveStorage($storage)
    {
        file_put_contents(static::STORAGE_PATH, '<?php
            $storage = '.var_export($storage, true).';
        ');
    }

    /**
     * Delete storage. Which means, the next compareStep() call starts over.
     *
     * @param bool $fullDelete Whether to delete all storage. Usually it's
     *                         sufficient to delete data related to the local
     *                         installation, only.
     *
     * @since 1.0.0
     */
    public static function deleteStorage($fullDelete = true)
    {
        if (file_exists(static::STORAGE_PATH)) {
            if ($fullDelete || ! is_readable(static::STORAGE_PATH)) {
                unlink(static::STORAGE_PATH);
            } else {
                $storage = [];
                require static::STORAGE_PATH;

                static::deleteNonpermanentStorageItems($storage);

                static::saveStorage($storage);
            }
        }
    }

    /**
     * Delete non-permanent items for the storage set.
     *
     * @param array $storage The storage set to deal with.
     *
     * @since 1.0.0
     */
    private static function deleteNonpermanentStorageItems(&$storage)
    {
        foreach (array_keys($storage) as $key) {
            $keep = false;
            foreach (static::STORAGE_PERMANENT_ITEMS as $filter) {
                if (preg_match($filter, $key)) {
                    $keep = true;
                }
            }
            if ( ! $keep) {
                unset($storage[$key]);
            }
        }
    }

    /**
     * Get translation for a given text.
     *
     * @param string $string String to translate.
     *
     * @return string Translation.
     *
     * @since 1.0.0
     */
    protected function l($string)
    {
        return Translate::getModuleTranslation('gitupdater', $string,
                                               'gitupdater');
    }

    /**
     * Download a list of files for a given version from api.thirtybees.com and
     * form it into a proper PHP array.
     *
     * For efficiency (a response can easily contain 10,000 lines), the
     * returned array contains just one key-value pair for each entry, path
     * and Git (SHA1) hash: ['<path>' => '<hash>']. Permissions get ignored,
     * because all files should have 644 permissions.
     *
     * @param string $version List for this version.
     *
     * @return bool|array Fileset, or false on failure.
     *
     * @since 1.0.0
     */
    protected function downloadFileList($version)
    {
        $response = false;
        $guzzle = new \GuzzleHttp\Client([
            'base_uri'    => AdminGitUpdaterController::API_URL,
            'verify'      => _PS_TOOL_DIR_.'cacert.pem',
            'timeout'     => 20,
        ]);
        try {
            $response = $guzzle->post('installationmaster.php', [
                'form_params' => [
                    'listrev' => $version,
                ],
            ])->getBody();
        } catch (Exception $e) {
            return false;
        }

        $fileList = false;
        if ($response) {
            $fileList = [];

            $adminDir = false;
            if (defined('_PS_ADMIN_DIR_')) {
                $adminDir = str_replace(_PS_ROOT_DIR_, '', _PS_ADMIN_DIR_);
                $adminDir = trim($adminDir, '/').'/';
            }

            foreach (json_decode($response) as $line) {
                // An incoming line is like '<permissions> blob <sha1>\t<path>'.
                // Use explode limits, to allow spaces in the last field.
                $fields = explode(' ', $line, 3);
                $line = $fields[2];
                $fields = explode("\t", $line, 2);
                $path = $fields[1];
                $hash = $fields[0];

                $keep = true;
                foreach (static::RELEASE_PATH_FILTER as $filter) {
                    if (preg_match($filter, $path)) {
                        $keep = false;
                        break;
                    }
                }
                if ($keep) {
                    foreach (static::KEEP_PATH_FILTER as $filter) {
                        if (preg_match($filter, $path)) {
                            $keep = false;
                            break;
                        }
                    }
                }

                if ($keep) {
                    if ($adminDir) {
                        $path = preg_replace('/^admin\//', $adminDir, $path);
                    }

                    $fileList[$path] = $hash;
                }
            }
        }

        return $fileList;
    }

    /**
     * Extract top level directories from a list of file paths. Purpose is
     * to allow splitting searches through the entire installation into smaller
     * chunks. Always present is the root directory, '.'.
     *
     * @param array $fileList List of file paths like documented with
     *                        downloadFileList().
     *
     * @return array List of paths. No failure expected.
     *
     * @since 1.0.0
     */
    protected function topLevelDirs($fileList)
    {
        $topLevelDirs = ['.'];

        foreach ($fileList as $path => $hash) {
            $slashPos = strpos($path, '/');

            // Ignore files at root level.
            if ($slashPos === false) {
                continue;
            }

            $topLevelDir = substr($path, 0, $slashPos);

            // vendor/ is huge, so take a second level.
            if ($topLevelDir === 'vendor') {
                $slashPos = strpos($path, '/', $slashPos + 1);
                if ($slashPos) {
                    $topLevelDir = substr($path, 0, $slashPos);
                }
            }

            if ( ! in_array($topLevelDir, $topLevelDirs)) {
                $topLevelDirs[] = $topLevelDir;
            }
        }

        return $topLevelDirs;
    }

    /**
     * Search installed files in a directory recursively and return a list
     * like specified in downloadFileList(), i.e. with Git hashes. This method
     * is expected to be called multiple times with distinct directories.
     *
     * Directories '.' and 'vendor' get searched not recursively. Note that
     * subdirectories of 'vendor' get searched as well, recursively.
     *
     * On return, $storage['installationlist'] exists and contains (additional)
     * paths to local files subject to verification.
     *
     * @param array  $storage All the data collected so far. This is expected
     *                        to be valid.
     * @param string $dir     Directory to search.
     *
     * @return array List of files. No failure expected, a not existing
     *               directory returns an empty array.
     *
     * @since 1.0.0
     */
    protected function searchInstallation(&$storage, $dir)
    {
        if ( ! array_key_exists('installationlist', $storage)) {
            $storage['installationlist'] = [];
        }
        $targetList = $storage['filelist-'.$storage['versionTarget']];
        $originList = $storage['filelist-'.$storage['versionOrigin']];

        $oldCwd = getcwd();
        chdir(_PS_ROOT_DIR_);

        if (is_dir($dir)) {
            if (in_array($dir, ['.', 'vendor'])) {
                $iterator = new DirectoryIterator($dir);
            } else {
                $iterator = new RecursiveIteratorIterator(
                                new RecursiveDirectoryIterator($dir)
                            );
            }

            foreach ($iterator as $fileInfo) {
                $path = $fileInfo->getPathname();
                if (in_array(basename($path), ['.', '..'])
                    || is_dir($path)) {
                    continue;
                }
                // Strip leading './'.
                $path = preg_replace('/(^.\/)/', '', $path);

                $keep = true;
                if ( ! array_key_exists($path, $targetList)
                    && ! array_key_exists($path, $originList)) {
                    foreach (static::INSTALLATION_PATH_FILTER as $filter) {
                        if (preg_match($filter, $path)) {
                            $keep = false;
                            break;
                        }
                    }
                }
                if ($keep) {
                    foreach (static::KEEP_PATH_FILTER as $filter) {
                        if (preg_match($filter, $path)) {
                            $keep = false;
                            break;
                        }
                    }
                }

                if ($keep) {
                    $content = file_get_contents($path);
                    $hash = sha1('blob '.strlen($content)."\0".$content);

                    $storage['installationlist'][$path] = $hash;
                }
            }
        }

        chdir($oldCwd);
    }

    /**
     * Calculate all the changes between the selected version and the current
     * installation.
     *
     * On return, $storage['changeset'] exists and contains an array of the
     * following format:
     *
     *               [
     *                   'change' => [
     *                       '<path>' => <manual>,
     *                       [...],
     *                   ],
     *                   'add' => [
     *                       '<path>' => <manual>,
     *                       [...],
     *                   ],
     *                   'remove' => [
     *                       '<path>' => <manual>,
     *                       [...],
     *                   ],
     *                   'obsolete' => [
     *                       '<path>' => <manual>,
     *                       [...],
     *                   ],
     *               ]
     *
     *               Where <path> is the path of the file and <manual> is a
     *               boolean indicating whether a change/add/remove overwrites
     *               manual edits.
     *
     * @param array  $storage All the data collected so far. This is expected
     *                        to be valid an complete.
     *
     * @since 1.0.0
     */
    protected function calculateChanges(&$storage)
    {
        $changeList   = [];
        $addList      = [];
        $removeList   = [];
        $obsoleteList = [];

        $targetList = $storage['filelist-'.$storage['versionTarget']];
        $installedList = $storage['installationlist'];
        $originList = $storage['filelist-'.$storage['versionOrigin']];

        foreach ($targetList as $path => $hash) {
            if (array_key_exists($path, $installedList)) {
                // Files to change are all files in the target version not
                // matching the installed file.
                if ($installedList[$path] !== $hash) {
                    $manual = false;
                    if (array_key_exists($path, $originList)
                        && $installedList[$path] !== $originList[$path]) {
                        $manual = true;
                    }
                    $changeList[$path] = $manual;
                } // else the file matches already.
            } else {
                // Files to add are all files in the target version not
                // existing locally.
                $addList[$path] = false;
            }
        }

        foreach ($installedList as $path => $hash) {
            if ( ! array_key_exists($path, $targetList)) {
                if (array_key_exists($path, $originList)) {
                    // Files to remove are all files not in the target version,
                    // but in the original version.
                    $manual = false;
                    if ($originList[$path] !== $hash) {
                        $manual = true;
                    }
                    $removeList[$path] = $manual;
                } else {
                    // Obsolete files are all files existing locally, but
                    // neither in the target nor in the original version.
                    $obsoleteList[$path] = true;
                }
            } // else handled above already.
        }

        $storage['changeset'] = [
            'change'    => $changeList,
            'add'       => $addList,
            'remove'    => $removeList,
            'obsolete'  => $obsoleteList,
        ];
    }
}
